/*********************************************************************
 *
 * File:         sessions.c
 *
 * Description:  NPAC Conferencing System API functions for session 
 *               management 
 *
 * Author:       Tomasz Stachowiak, NPAC
 *
 * Date:         10-14-96
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

#include <netdb.h>
#include <unistd.h>

#include "common.h"

Participant member[MAX_MEMBERS];
int memberActive[MAX_MEMBERS];
int localHandle;

void ncsEngineMessageHandle(short tag, void *msg, int len);
void ncsReadEngineMessage();


void (* inviteCallback)(void *);
void (* timeoutCallback)(void *);
void (* connectedCallback)(void *);
void (* acceptCallback)(void *);
void (* rejectCallback)(void *);
void (* deleteMemberCallback)(void *);
void (* addMemberCallback)(void *);
void (* startAppCallback)(void *);
void (* serverUsersCallback)(void *);

ManagerData managerData;


void InitMembers()
{
  int i;

  for(i=0; i<MAX_MEMBERS; i++)
    memberActive[i] = 0;
}

void InitCallbacks()
{
  inviteCallback = NULL;
  timeoutCallback = NULL;
  deleteMemberCallback = NULL;
  addMemberCallback = NULL;
  startAppCallback = NULL;
  serverUsersCallback = NULL;
  connectedCallback = NULL;
  acceptCallback = NULL;
  rejectCallback = NULL;
}

void AddMember(char *data)
{
  Participant *newmember;
  NCSparticipant part;

  newmember = (Participant *)data;
  memberActive[newmember->handle] = 1;
  memcpy(&member[newmember->handle], newmember, sizeof(Participant));

  strcpy(part.user, newmember->user);
  if(getHostName(newmember->host, part.host) > 0)
    {
      PRINTF("Manager: Member added locally %s@%s handle %d\n", part.user, part.host, newmember->handle);
      if(addMemberCallback != NULL)
	addMemberCallback(&part);
    }
  else
    PRINTF("Manager: Host unknown\n");
}

void InvitationReceived(char *data)
{
  Participant *member;
  NCSparticipant part;

  member = (Participant *)data;
  strcpy(part.user, member->user);
  if(getHostName(member->host, part.host) > 0)
    {
      if(inviteCallback != NULL)
	inviteCallback(&part);
    }
  else
    PRINTF("Manager: Host unknown\n");
}

void InvitationAccepted(char *data)
{
  Participant *member;
  NCSparticipant part;

  member = (Participant *)data;
  strcpy(part.user, member->user);
  if(getHostName(member->host, part.host) > 0)
    {
      if(acceptCallback != NULL)
	acceptCallback(&part);
    }
  else
    PRINTF("Manager: Host unknown\n");
}

void InvitationRejected(char *data)
{
  Participant *member;
  NCSparticipant part;

  member = (Participant *)data;
  strcpy(part.user, member->user);
  if(getHostName(member->host, part.host) > 0)
    {
      if(rejectCallback != NULL)
	rejectCallback(&part);
    }
  else
    PRINTF("Manager: Host unknown\n");
}

void InvitationTimeout(char *data)
{
  Participant *member;
  NCSparticipant part;

  member = (Participant *)data;
  strcpy(part.user, member->user);
  if(getHostName(member->host, part.host) > 0)
    {
      if(timeoutCallback != NULL)
	timeoutCallback(&part);
    }
  else
    PRINTF("Manager: Host unknown\n");
}

void InvitationDelivered(char *data)
{
  Participant *member;
  NCSparticipant part;

  member = (Participant *)data;
  strcpy(part.user, member->user);
  if(getHostName(member->host, part.host) > 0)
    {
      if(connectedCallback != NULL)
	connectedCallback(&part);
    }
  else
    PRINTF("Manager: Host unknown\n");
}

void DeleteMember(char *data)
{
  NCSparticipant part;
  int *no;

  signal(SIGCHLD, SIG_IGN);

  no = (int *)data;
  memberActive[*no] = 0;

  strcpy(part.user, member[*no].user);
  if(getHostName(member[*no].host, part.host) > 0)
    {
      if(deleteMemberCallback != NULL)
	deleteMemberCallback(&part);
    }
  else
    PRINTF("Manager: Host unknown\n");
}


int ncsSessionInitialize()
{
  static int first = 1;
  long hostid;

  if ((managerData.socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
      fprintf(stderr, "Manager: can't open manage socket\n");
      return 0;
    }

  /*
  bzero(&managerData.addr, sizeof(managerData.addr));
  managerData.addr.sin_family = AF_INET;
  managerData.addr.sin_addr.s_addr = htonl(INADDR_ANY);
  managerData.addr.sin_port        = htons(MANAGER_PORT);

  if(bind(managerData.socket, &managerData.addr, sizeof(managerData.addr)) < 0)
    {
      fprintf(stderr, "Manager: can't bind daemon socket\n");
      return 0;
    }
    */

  bzero(&managerData.daemon_addr, sizeof(managerData.daemon_addr));
  managerData.daemon_addr.sin_family = AF_INET;
  hostid = gethostid();
  managerData.daemon_addr.sin_addr.s_addr = htonl(hostid);
  managerData.daemon_addr.sin_port        = htons(DAEMON_PORT);

  if(connect(managerData.socket, &managerData.daemon_addr, sizeof(managerData.daemon_addr)) < 0)
    { 
      fprintf(stderr, "Manager: can not connect to engine\n");
      exit(0);
    }

  ncsEngineRequest(reqNewLocalSessionInit, NULL, 0);

  InitMembers();
  if(first)
    {
      InitCallbacks();
      first = 0;
    }

  serverAdd();

  return 1;
}

void ncsSessionCreate()
{
  ncsEngineRequest(reqLocalSessionCreate, NULL, 0);
  localHandle = 0;     /* what if session already created?!!! */
}


/* Registers working procedure running in the Xt background, 
   processing incoming control messages */
void ncsSessionRegisterWithXt(/*XtApp app*/)
{
}

/* Returns file descriptor to session control socket */
int ncsSessionGetFD(fd_set *set)
{
  FD_ZERO(set);
  FD_SET(managerData.socket, set);
  return managerData.socket;
}

/* Processes control messages coming to the control socket */
void ncsProcessIncoming(fd_set *set)
{
  void *msg;
  int n, msg_len;
  struct sockaddr_in addr;
  int addr_size;
  
  addr_size = sizeof(addr);
  if(FD_ISSET(managerData.socket, set))
    {
      ncsReadEngineMessage();
    }
}



void ncsReadEngineMessage()
{
  char *msg;
  MsgHeader header;
  int err;

  err = readMsgHeader(managerData.socket, &header);
  if(err < 0)
    {
      fprintf(stderr, "Manager: data reading error\n");
      return;
    }
  if(err == 0)
    {
      fprintf(stderr, "Manager: engine killed, exitting...\n");
      exit(0);
    }
  if(header.dataLength > 0)
    {
      msg = (char *)malloc(header.dataLength);
      err = readAll(managerData.socket, msg, header.dataLength);
      if(err < 0)
	{
	  fprintf(stderr, "Manager: data reading error\n");
	  return;
	}
      if(err == 0)
	{
	  fprintf(stderr, "Manager: engine killed, exitting...\n");
	  exit(0);
	}
    }
  else
    msg = NULL;
  
  ncsEngineMessageHandle(header.tag, msg, header.dataLength);
  
  free(msg);
}

void ncsEngineMessageHandle(short tag, void *msg, int len)
{
  InviteData *info;

  info = (InviteData *)msg;
  switch((LocalTag)tag)
    {
    case indLocalSessionInvited:
      PRINTF("Manager: Invitation received\n");
      InvitationReceived(msg);
      break;

    case indLocalSessionAccept:
      PRINTF("Manager: Invitation accepted\n");
      InvitationAccepted(msg);
      break;

    case indLocalSessionReject:
      PRINTF("Manager: Invitation rejected\n");
      InvitationRejected(msg);
      break;
      
    case indLocalSessionMembers:
      break;

    case indLocalSessionAddMember:
      PRINTF("Manager: add member\n");
      AddMember(msg);
      break;

    case indLocalSessionDeleteMember:
      PRINTF("Manager: delete member\n");
      DeleteMember(msg);
      break;

    case indLocalSessionLocalMember:
      PRINTF("Manager: local handle %d\n", *(int *)msg);
      localHandle = *(int *)msg;
      break;

    case indLocalSessionStartApp:
      PRINTF("Manager: start application\n");
      if(startAppCallback != NULL)
	startAppCallback(msg);
      break;

    case indLocalSessionTimeout:
      PRINTF("Manager: invitation timeout\n");
      InvitationTimeout(msg);
      break;

    case indLocalSessionConnected:
      PRINTF("Manager: invitation delivered\n");
      InvitationDelivered(msg);
      break;

    case indServerUsers:
      PRINTF("Manager: server users\n");
      serverUsersCallback(msg);
      break;

    default:
      fprintf(stderr, "Manager: wrong message tag %d\n", (LocalTag)tag);
    }
}

void ncsSessionAddMember(char *user, char *host)
{
  Participant part;

  strcpy(part.user, user);
  part.host = getHostID(host);
  PRINTF("Manager: invite member %s@%d\n", part.user, part.host);
  if(part.host != 0)
    ncsEngineRequest(reqLocalSessionAddMember, &part, sizeof(Participant));
  else
    PRINTF("Manager: wrong host name\n");
}

void ncsSessionAcceptInvitation(NCSparticipant *part)
{
  Participant member;

  strcpy(member.user, part->user);
  member.host = getHostID(part->host);
  if(member.host != 0)
    {
      PRINTF("Manager: join session %s@%s \n", part->user, part->host);
      ncsEngineRequest(reqLocalSessionJoin, &member, sizeof(Participant));
    }
  else
    PRINTF("Manager: unknown host %s@%s \n", part->user, part->host);
}

void ncsSessionRejectInvitation(NCSparticipant *part)
{
  Participant member;

  strcpy(member.user, part->user);
  member.host = getHostID(part->host);
  if(member.host != 0)
    {
      PRINTF("Manager: delete invitation from %s@%s \n", part->user, part->host);
      ncsEngineRequest(reqLocalSessionReject, &member, sizeof(Participant));
    }
  else
    PRINTF("Manager: unknown host %s@%s \n", part->user, part->host);
}

void ncsSessionAddCallback(short tag, void (*function))
{
  switch(tag)
    {
    case indLocalSessionInvited:
      inviteCallback = function;
      break;

    case indLocalSessionAddMember:
      addMemberCallback = function;
      break;

    case indLocalSessionDeleteMember:
      deleteMemberCallback = function;
      break;

    case indLocalSessionStartApp:
      startAppCallback = function;
      break;

    case indLocalSessionTimeout:
      timeoutCallback = function;
      break;

    case indLocalSessionConnected:
      connectedCallback = function;
      break;

    case indLocalSessionAccept:
      acceptCallback = function;
      break;

    case indLocalSessionReject:
      rejectCallback = function;
      break;

    case indServerUsers:
      serverUsersCallback = function;
      break;

    default:
      fprintf(stderr, "Manager: Unknown control callback type\n");
    }
}

void ncsSessionQuit()
{
  serverDel();
  ncsEngineRequest(reqLocalSessionQuit, NULL, 0);
  close(managerData.socket);
}

void ncsSessionQueryMembers()
{
  ncsEngineRequest(reqLocalSessionGetMembers, NULL, 0);
}


void ncsSessionSendToApp(int port, short type, char *msg, int len)
{
  AppData *data;
  int size;

  if(len == 0)
    {
      size = sizeof(AppData);
      data = malloc(size);
      data->data[0] = 0;
    }
  else
    {
      size = sizeof(AppData) + len - 1;
      data = malloc(size);
      memcpy(data->data, data, len);
    }
    

  data->port = port;
  PRINTF("Manager: Send to app port %d, msg size %d\n", data->port, size);
  data->tag = type;
  memcpy(data->data, msg, len);
  ncsEngineRequest(reqLocalSessionSendToApp, data, size);
  free(data);
}

void ncsSessionStartApp(int port)
{
  ncsEngineRequest(reqLocalSessionStartApp, &port, sizeof(int));
}

void ncsSessionReset()
{ 
  ncsEngineRequest(reqLocalSessionReset, NULL, 0);
}

int ncsSessionSaveInit()
{
  ncsEngineRequest(reqLocalSessionSaveInit, NULL, 0);
}

void ncsSessionSaveClose()
{
  ncsEngineRequest(reqLocalSessionSaveClose, NULL, 0);
}

int ncsSessionRestoreInit()
{
  ncsEngineRequest(reqLocalSessionRestoreInit, NULL, 0);
}

void ncsSessionRestoreClose()
{
  ncsEngineRequest(reqLocalSessionRestoreClose, NULL, 0);
}





